@@ -524,7 +524,16 @@ class DynamicCapability::Server: public Capability::Server {
 public:
   typedef DynamicCapability Serves;
 
+  struct Options {
+    bool allowCancellation = false;
+    // See the `allowCancellation` annotation defined in `c++.capnp`.
+    //
+    // This option applies to all calls made to this server object. The annotation in the schema
+    // is NOT used for dynamic servers.
+  };
+
   Server(InterfaceSchema schema): schema(schema) {}
+  Server(InterfaceSchema schema, Options options): schema(schema), options(options) {}
 
   virtual kj::Promise<void> call(InterfaceSchema::Method method,
                                  CallContext<DynamicStruct, DynamicStruct> context) = 0;
@@ -536,6 +545,7 @@ public:
 
 private:
   InterfaceSchema schema;
+  Options options;
 };
 
 template <>
@@ -584,7 +594,6 @@ public:
   Orphanage getResultsOrphanage(kj::Maybe<MessageSize> sizeHint = nullptr);
   template <typename SubParams>
   kj::Promise<void> tailCall(Request<SubParams, DynamicStruct>&& tailRequest);
-  void allowCancellation();
 
   StructSchema getParamsType() const { return paramType; }
   StructSchema getResultsType() const { return resultType; }
@@ -1658,9 +1667,6 @@ inline kj::Promise<void> CallContext<DynamicStruct, DynamicStruct>::tailCall(
     Request<SubParams, DynamicStruct>&& tailRequest) {
   return hook->tailCall(kj::mv(tailRequest.hook));
 }
-inline void CallContext<DynamicStruct, DynamicStruct>::allowCancellation() {
-  hook->allowCancellation();
-}
 
 template <>
 inline DynamicCapability::Client Capability::Client::castAs<DynamicCapability>(
@@ -1668,6 +1674,14 @@ inline DynamicCapability::Client Capability::Client::castAs<DynamicCapability>(
   return DynamicCapability::Client(schema, hook->addRef());
 }
 
+template <>
+inline DynamicCapability::Client CapabilityServerSet<DynamicCapability>::add(
+    kj::Own<DynamicCapability::Server>&& server) {
+  void* ptr = reinterpret_cast<void*>(server.get());
+  auto schema = server->getSchema();
+  return addInternal(kj::mv(server), ptr).castAs<DynamicCapability>(schema);
+}
+
 // -------------------------------------------------------------------
 
 template <typename T>
