@@ -212,6 +212,7 @@ public:
 
   kj::Promise<void> getCap(GetCapContext context) override;
   kj::Promise<void> getAnyCap(GetAnyCapContext context) override;
+  kj::Promise<void> getCapPipelineOnly(GetCapPipelineOnlyContext context) override;
 
 private:
   int& callCount;
@@ -277,6 +278,10 @@ public:
 
   kj::Promise<void> writeToFd(WriteToFdContext context) override;
 
+  kj::Promise<void> throwException(ThrowExceptionContext context) override;
+
+  kj::Promise<void> throwRemoteException(ThrowRemoteExceptionContext context) override;
+
 private:
   int& callCount;
   int& handleCount;
@@ -333,7 +338,6 @@ public:
   }
 
   kj::Promise<void> doStreamJ(DoStreamJContext context) override {
-    context.allowCancellation();
     jSum += context.getParams().getJ();
 
     if (jShouldThrow) {
