@@ -23,13 +23,16 @@
 
 #include <initializer_list>
 #include "array.h"
+#include "kj/common.h"
 #include <string.h>
 
 KJ_BEGIN_HEADER
 
 namespace kj {
   class StringPtr;
+  class LiteralStringConst;
   class String;
+  class ConstString;
 
   class StringTree;   // string-tree.h
 }
@@ -50,6 +53,8 @@ constexpr kj::StringPtr operator "" _kj(const char* str, size_t n);
 // string literal vs. one with _kj (assuming the compiler is able to optimize away strlen() on a
 // string literal).
 
+constexpr kj::LiteralStringConst operator "" _kjc(const char* str, size_t n);
+
 namespace kj {
 
 // Our STL string SFINAE trick does not work with GCC 4.7, but it works with Clang and GCC 4.8, so
@@ -76,6 +81,7 @@ public:
   inline StringPtr(const char* begin KJ_LIFETIMEBOUND, const char* end KJ_LIFETIMEBOUND): StringPtr(begin, end - begin) {}
   inline StringPtr(String&& value KJ_LIFETIMEBOUND) : StringPtr(value) {}
   inline StringPtr(const String& value KJ_LIFETIMEBOUND);
+  inline StringPtr(const ConstString& value KJ_LIFETIMEBOUND);
   StringPtr& operator=(String&& value) = delete;
   inline StringPtr& operator=(decltype(nullptr)) {
     content = ArrayPtr<const char>("", 1);
@@ -92,15 +98,20 @@ public:
 #endif
 
 #if KJ_COMPILER_SUPPORTS_STL_STRING_INTEROP
-  template <typename T, typename = decltype(instance<T>().c_str())>
-  inline StringPtr(const T& t KJ_LIFETIMEBOUND): StringPtr(t.c_str()) {}
-  // Allow implicit conversion from any class that has a c_str() method (namely, std::string).
+  template <
+    typename T,
+    typename = decltype(instance<T>().c_str()),
+    typename = decltype(instance<T>().size())>
+  inline StringPtr(const T& t KJ_LIFETIMEBOUND): StringPtr(t.c_str(), t.size()) {}
+  // Allow implicit conversion from any class that has a c_str() and a size() method (namely, std::string).
   // We use a template trick to detect std::string in order to avoid including the header for
   // those who don't want it.
-
-  template <typename T, typename = decltype(instance<T>().c_str())>
-  inline operator T() const { return cStr(); }
-  // Allow implicit conversion to any class that has a c_str() method (namely, std::string).
+  template <
+    typename T,
+    typename = decltype(instance<T>().c_str()),
+    typename = decltype(instance<T>().size())>
+  inline operator T() const { return {cStr(), size()}; }
+  // Allow implicit conversion to any class that has a c_str() method and a size() method (namely, std::string).
   // We use a template trick to detect std::string in order to avoid including the header for
   // those who don't want it.
 #endif
@@ -122,10 +133,14 @@ public:
   inline constexpr const char* end() const { return content.end() - 1; }
 
   inline constexpr bool operator==(decltype(nullptr)) const { return content.size() <= 1; }
+#if !__cpp_impl_three_way_comparison
   inline constexpr bool operator!=(decltype(nullptr)) const { return content.size() > 1; }
+#endif
 
   inline bool operator==(const StringPtr& other) const;
+#if !__cpp_impl_three_way_comparison
   inline bool operator!=(const StringPtr& other) const { return !(*this == other); }
+#endif
   inline bool operator< (const StringPtr& other) const;
   inline bool operator> (const StringPtr& other) const { return other < *this; }
   inline bool operator<=(const StringPtr& other) const { return !(other < *this); }
@@ -136,11 +151,11 @@ public:
   // A string slice is only NUL-terminated if it is a suffix, so slice() has a one-parameter
   // version that assumes end = size().
 
-  inline bool startsWith(const StringPtr& other) const;
-  inline bool endsWith(const StringPtr& other) const;
+  inline bool startsWith(const StringPtr& other) const { return asArray().startsWith(other);}
+  inline bool endsWith(const StringPtr& other) const { return asArray().endsWith(other); }
 
-  inline Maybe<size_t> findFirst(char c) const;
-  inline Maybe<size_t> findLast(char c) const;
+  inline Maybe<size_t> findFirst(char c) const { return asArray().findFirst(c); }
+  inline Maybe<size_t> findLast(char c) const { return asArray().findLast(c); }
 
   template <typename T>
   T parseAs() const;
@@ -149,13 +164,22 @@ public:
   // Integer numbers prefixed by "0" are parsed in base 10 (unlike strtoi with base 0).
   // Overflowed integer numbers throw exception.
   // Overflowed floating numbers return inf.
+  template <typename T>
+  Maybe<T> tryParseAs() const;
+  // Same as parseAs, but rather than throwing an exception we return NULL.
+
+  template <typename... Attachments>
+  ConstString attach(Attachments&&... attachments) const KJ_WARN_UNUSED_RESULT;
+  ConstString attach() const KJ_WARN_UNUSED_RESULT;
+  // Like ArrayPtr<T>::attach(), but instead promotes a StringPtr into a ConstString. Generally the
+  // attachment should be an object that somehow owns the String that the StringPtr is pointing at.
 
 private:
   inline explicit constexpr StringPtr(ArrayPtr<const char> content): content(content) {}
+  friend constexpr StringPtr (::operator "" _kj)(const char* str, size_t n);
+  friend class LiteralStringConst;
 
   ArrayPtr<const char> content;
-
-  friend constexpr kj::StringPtr (::operator "" _kj)(const char* str, size_t n);
   friend class SourceLocation;
 };
 
@@ -178,6 +202,29 @@ template <> unsigned long long StringPtr::parseAs<unsigned long long>() const;
 template <> float StringPtr::parseAs<float>() const;
 template <> double StringPtr::parseAs<double>() const;
 
+template <> Maybe<char> StringPtr::tryParseAs<char>() const;
+template <> Maybe<signed char> StringPtr::tryParseAs<signed char>() const;
+template <> Maybe<unsigned char> StringPtr::tryParseAs<unsigned char>() const;
+template <> Maybe<short> StringPtr::tryParseAs<short>() const;
+template <> Maybe<unsigned short> StringPtr::tryParseAs<unsigned short>() const;
+template <> Maybe<int> StringPtr::tryParseAs<int>() const;
+template <> Maybe<unsigned> StringPtr::tryParseAs<unsigned>() const;
+template <> Maybe<long> StringPtr::tryParseAs<long>() const;
+template <> Maybe<unsigned long> StringPtr::tryParseAs<unsigned long>() const;
+template <> Maybe<long long> StringPtr::tryParseAs<long long>() const;
+template <> Maybe<unsigned long long> StringPtr::tryParseAs<unsigned long long>() const;
+template <> Maybe<float> StringPtr::tryParseAs<float>() const;
+template <> Maybe<double> StringPtr::tryParseAs<double>() const;
+
+class LiteralStringConst: public StringPtr {
+public:
+  inline operator ConstString() const;
+
+private:
+  inline explicit constexpr LiteralStringConst(ArrayPtr<const char> content): StringPtr(content) {}
+  friend constexpr LiteralStringConst (::operator "" _kjc)(const char* str, size_t n);
+};
+
 // =======================================================================================
 // String -- A NUL-terminated Array<char> containing UTF-8 text.
 //
@@ -231,14 +278,18 @@ public:
   inline bool operator!=(decltype(nullptr)) const { return content.size() > 1; }
 
   inline bool operator==(const StringPtr& other) const { return StringPtr(*this) == other; }
+#if !__cpp_impl_three_way_comparison
   inline bool operator!=(const StringPtr& other) const { return StringPtr(*this) != other; }
+#endif
   inline bool operator< (const StringPtr& other) const { return StringPtr(*this) <  other; }
   inline bool operator> (const StringPtr& other) const { return StringPtr(*this) >  other; }
   inline bool operator<=(const StringPtr& other) const { return StringPtr(*this) <= other; }
   inline bool operator>=(const StringPtr& other) const { return StringPtr(*this) >= other; }
 
   inline bool operator==(const String& other) const { return StringPtr(*this) == StringPtr(other); }
+#if !__cpp_impl_three_way_comparison
   inline bool operator!=(const String& other) const { return StringPtr(*this) != StringPtr(other); }
+#endif
   inline bool operator< (const String& other) const { return StringPtr(*this) <  StringPtr(other); }
   inline bool operator> (const String& other) const { return StringPtr(*this) >  StringPtr(other); }
   inline bool operator<=(const String& other) const { return StringPtr(*this) <= StringPtr(other); }
@@ -247,8 +298,17 @@ public:
   // comparisons between two strings are ambiguous. (Clang turns this into a warning,
   // -Wambiguous-reversed-operator, due to the stupidity...)
 
-  inline bool startsWith(const StringPtr& other) const { return StringPtr(*this).startsWith(other);}
-  inline bool endsWith(const StringPtr& other) const { return StringPtr(*this).endsWith(other); }
+  inline bool operator==(const ConstString& other) const { return StringPtr(*this) == StringPtr(other); }
+#if !__cpp_impl_three_way_comparison
+  inline bool operator!=(const ConstString& other) const { return StringPtr(*this) != StringPtr(other); }
+#endif
+  inline bool operator< (const ConstString& other) const { return StringPtr(*this) <  StringPtr(other); }
+  inline bool operator> (const ConstString& other) const { return StringPtr(*this) >  StringPtr(other); }
+  inline bool operator<=(const ConstString& other) const { return StringPtr(*this) <= StringPtr(other); }
+  inline bool operator>=(const ConstString& other) const { return StringPtr(*this) >= StringPtr(other); }
+
+  inline bool startsWith(const StringPtr& other) const { return asArray().startsWith(other);}
+  inline bool endsWith(const StringPtr& other) const { return asArray().endsWith(other); }
 
   inline StringPtr slice(size_t start) const KJ_LIFETIMEBOUND {
     return StringPtr(*this).slice(start);
@@ -257,17 +317,120 @@ public:
     return StringPtr(*this).slice(start, end);
   }
 
-  inline Maybe<size_t> findFirst(char c) const { return StringPtr(*this).findFirst(c); }
-  inline Maybe<size_t> findLast(char c) const { return StringPtr(*this).findLast(c); }
+  inline Maybe<size_t> findFirst(char c) const { return asArray().findFirst(c); }
+  inline Maybe<size_t> findLast(char c) const { return asArray().findLast(c); }
 
   template <typename T>
   T parseAs() const { return StringPtr(*this).parseAs<T>(); }
   // Parse as number
 
+  template <typename T>
+  Maybe<T> tryParseAs() const { return StringPtr(*this).tryParseAs<T>(); }
+
 private:
   Array<char> content;
 };
 
+// =======================================================================================
+// ConstString -- Same as String, but the backing buffer is const.
+//
+// This has the useful property that it can reference a string literal without allocating
+// a copy. Any String can also convert (by move) to ConstString, transferring ownership of
+// the buffer.
+
+class ConstString {
+public:
+  ConstString() = default;
+  inline ConstString(decltype(nullptr)): content(nullptr) {}
+  inline ConstString(const char* value, size_t size, const ArrayDisposer& disposer);
+  // Does not copy.  `size` does not include NUL terminator, but `value` must be NUL-terminated.
+  inline explicit ConstString(Array<const char> buffer);
+  // Does not copy.  Requires `buffer` ends with `\0`.
+  inline explicit ConstString(String&& string): content(string.releaseArray()) {}
+  // Does not copy. Ownership is transfered.
+
+  inline operator ArrayPtr<const char>() const KJ_LIFETIMEBOUND;
+  inline ArrayPtr<const char> asArray() const KJ_LIFETIMEBOUND;
+  inline ArrayPtr<const byte> asBytes() const KJ_LIFETIMEBOUND { return asArray().asBytes(); }
+  // Result does not include NUL terminator.
+
+  inline StringPtr asPtr() const KJ_LIFETIMEBOUND {
+    // Convenience operator to return a StringPtr.
+    return StringPtr{*this};
+  }
+
+  inline Array<const char> releaseArray() { return kj::mv(content); }
+  // Disowns the backing array (which includes the NUL terminator) and returns it. The ConstString value
+  // is clobbered (as if moved away).
+
+  inline const char* cStr() const KJ_LIFETIMEBOUND;
+
+  inline size_t size() const;
+  // Result does not include NUL terminator.
+
+  inline char operator[](size_t index) const;
+  inline char& operator[](size_t index) KJ_LIFETIMEBOUND;
+
+  inline const char* begin() const KJ_LIFETIMEBOUND;
+  inline const char* end() const KJ_LIFETIMEBOUND;
+
+  inline bool operator==(decltype(nullptr)) const { return content.size() <= 1; }
+  inline bool operator!=(decltype(nullptr)) const { return content.size() > 1; }
+
+  inline bool operator==(const StringPtr& other) const { return StringPtr(*this) == other; }
+#if !__cpp_impl_three_way_comparison
+  inline bool operator!=(const StringPtr& other) const { return StringPtr(*this) != other; }
+#endif
+  inline bool operator< (const StringPtr& other) const { return StringPtr(*this) <  other; }
+  inline bool operator> (const StringPtr& other) const { return StringPtr(*this) >  other; }
+  inline bool operator<=(const StringPtr& other) const { return StringPtr(*this) <= other; }
+  inline bool operator>=(const StringPtr& other) const { return StringPtr(*this) >= other; }
+
+  inline bool operator==(const String& other) const { return StringPtr(*this) == StringPtr(other); }
+#if !__cpp_impl_three_way_comparison
+  inline bool operator!=(const String& other) const { return StringPtr(*this) != StringPtr(other); }
+#endif
+  inline bool operator< (const String& other) const { return StringPtr(*this) <  StringPtr(other); }
+  inline bool operator> (const String& other) const { return StringPtr(*this) >  StringPtr(other); }
+  inline bool operator<=(const String& other) const { return StringPtr(*this) <= StringPtr(other); }
+  inline bool operator>=(const String& other) const { return StringPtr(*this) >= StringPtr(other); }
+
+  inline bool operator==(const ConstString& other) const { return StringPtr(*this) == StringPtr(other); }
+#if !__cpp_impl_three_way_comparison
+  inline bool operator!=(const ConstString& other) const { return StringPtr(*this) != StringPtr(other); }
+#endif
+  inline bool operator< (const ConstString& other) const { return StringPtr(*this) <  StringPtr(other); }
+  inline bool operator> (const ConstString& other) const { return StringPtr(*this) >  StringPtr(other); }
+  inline bool operator<=(const ConstString& other) const { return StringPtr(*this) <= StringPtr(other); }
+  inline bool operator>=(const ConstString& other) const { return StringPtr(*this) >= StringPtr(other); }
+  // Note that if we don't overload for `const ConstString&` specifically, then C++20 will decide that
+  // comparisons between two strings are ambiguous. (Clang turns this into a warning,
+  // -Wambiguous-reversed-operator, due to the stupidity...)
+
+  inline bool startsWith(const StringPtr& other) const { return asArray().startsWith(other);}
+  inline bool endsWith(const StringPtr& other) const { return asArray().endsWith(other); }
+
+  inline StringPtr slice(size_t start) const KJ_LIFETIMEBOUND {
+    return StringPtr(*this).slice(start);
+  }
+  inline ArrayPtr<const char> slice(size_t start, size_t end) const KJ_LIFETIMEBOUND {
+    return StringPtr(*this).slice(start, end);
+  }
+
+  inline Maybe<size_t> findFirst(char c) const { return asArray().findFirst(c); }
+  inline Maybe<size_t> findLast(char c) const { return asArray().findLast(c); }
+
+  template <typename T>
+  T parseAs() const { return StringPtr(*this).parseAs<T>(); }
+  // Parse as number
+
+  template <typename T>
+  Maybe<T> tryParseAs() const { return StringPtr(*this).tryParseAs<T>(); }
+
+private:
+  Array<const char> content;
+};
+
 #if !__cpp_impl_three_way_comparison
 inline bool operator==(const char* a, const String& b) { return b == a; }
 inline bool operator!=(const char* a, const String& b) { return b != a; }
@@ -397,6 +560,7 @@ struct Stringifier {
     return s.asArray();
   }
   inline ArrayPtr<const char> operator*(const StringPtr& s) const { return s.asArray(); }
+  inline ArrayPtr<const char> operator*(const ConstString& s) const { return s.asArray(); }
 
   inline Range<char> operator*(const Range<char>& r) const { return r; }
   inline Repeat<char> operator*(const Repeat<char>& r) const { return r; }
@@ -544,6 +708,7 @@ inline _::Delimited<ArrayPtr<const T>> operator*(const _::Stringifier&, const Ar
 // Inline implementation details.
 
 inline StringPtr::StringPtr(const String& value): content(value.cStr(), value.size() + 1) {}
+inline StringPtr::StringPtr(const ConstString& value): content(value.cStr(), value.size() + 1) {}
 
 inline constexpr StringPtr::operator ArrayPtr<const char>() const {
   return ArrayPtr<const char>(content.begin(), content.size() - 1);
@@ -572,31 +737,18 @@ inline ArrayPtr<const char> StringPtr::slice(size_t start, size_t end) const {
   return content.slice(start, end);
 }
 
-inline bool StringPtr::startsWith(const StringPtr& other) const {
-  return other.content.size() <= content.size() &&
-      memcmp(content.begin(), other.content.begin(), other.size()) == 0;
-}
-inline bool StringPtr::endsWith(const StringPtr& other) const {
-  return other.content.size() <= content.size() &&
-      memcmp(end() - other.size(), other.content.begin(), other.size()) == 0;
+inline LiteralStringConst::operator ConstString() const {
+  return ConstString(begin(), size(), NullArrayDisposer::instance);
 }
 
-inline Maybe<size_t> StringPtr::findFirst(char c) const {
-  const char* pos = reinterpret_cast<const char*>(memchr(content.begin(), c, size()));
-  if (pos == nullptr) {
-    return nullptr;
-  } else {
-    return pos - content.begin();
-  }
+inline ConstString StringPtr::attach() const {
+  // This is meant as a roundabout way to make a ConstString from a StringPtr
+  return ConstString(begin(), size(), NullArrayDisposer::instance);
 }
 
-inline Maybe<size_t> StringPtr::findLast(char c) const {
-  for (size_t i = size(); i > 0; --i) {
-    if (content[i-1] == c) {
-      return i-1;
-    }
-  }
-  return nullptr;
+template <typename... Attachments>
+inline ConstString StringPtr::attach(Attachments&&... attachments) const {
+  return ConstString { .content = content.attach(kj::fwd<Attachments>(attachments)...) };
 }
 
 inline String::operator ArrayPtr<char>() {
@@ -605,6 +757,9 @@ inline String::operator ArrayPtr<char>() {
 inline String::operator ArrayPtr<const char>() const {
   return content == nullptr ? ArrayPtr<const char>(nullptr) : content.slice(0, content.size() - 1);
 }
+inline ConstString::operator ArrayPtr<const char>() const {
+  return content == nullptr ? ArrayPtr<const char>(nullptr) : content.slice(0, content.size() - 1);
+}
 
 inline ArrayPtr<char> String::asArray() {
   return content == nullptr ? ArrayPtr<char>(nullptr) : content.slice(0, content.size() - 1);
@@ -612,27 +767,42 @@ inline ArrayPtr<char> String::asArray() {
 inline ArrayPtr<const char> String::asArray() const {
   return content == nullptr ? ArrayPtr<const char>(nullptr) : content.slice(0, content.size() - 1);
 }
+inline ArrayPtr<const char> ConstString::asArray() const {
+  return content == nullptr ? ArrayPtr<const char>(nullptr) : content.slice(0, content.size() - 1);
+}
 
 inline const char* String::cStr() const { return content == nullptr ? "" : content.begin(); }
+inline const char* ConstString::cStr() const { return content == nullptr ? "" : content.begin(); }
 
 inline size_t String::size() const { return content == nullptr ? 0 : content.size() - 1; }
+inline size_t ConstString::size() const { return content == nullptr ? 0 : content.size() - 1; }
 
 inline char String::operator[](size_t index) const { return content[index]; }
 inline char& String::operator[](size_t index) { return content[index]; }
+inline char ConstString::operator[](size_t index) const { return content[index]; }
 
 inline char* String::begin() { return content == nullptr ? nullptr : content.begin(); }
 inline char* String::end() { return content == nullptr ? nullptr : content.end() - 1; }
 inline const char* String::begin() const { return content == nullptr ? nullptr : content.begin(); }
 inline const char* String::end() const { return content == nullptr ? nullptr : content.end() - 1; }
+inline const char* ConstString::begin() const { return content == nullptr ? nullptr : content.begin(); }
+inline const char* ConstString::end() const { return content == nullptr ? nullptr : content.end() - 1; }
 
 inline String::String(char* value, size_t size, const ArrayDisposer& disposer)
     : content(value, size + 1, disposer) {
   KJ_IREQUIRE(value[size] == '\0', "String must be NUL-terminated.");
 }
+inline ConstString::ConstString(const char* value, size_t size, const ArrayDisposer& disposer)
+    : content(value, size + 1, disposer) {
+  KJ_IREQUIRE(value[size] == '\0', "String must be NUL-terminated.");
+}
 
 inline String::String(Array<char> buffer): content(kj::mv(buffer)) {
   KJ_IREQUIRE(content.size() > 0 && content.back() == '\0', "String must be NUL-terminated.");
 }
+inline ConstString::ConstString(Array<const char> buffer): content(kj::mv(buffer)) {
+  KJ_IREQUIRE(content.size() > 0 && content.back() == '\0', "String must be NUL-terminated.");
+}
 
 inline String heapString(const char* value) {
   return heapString(value, strlen(value));
@@ -758,4 +928,8 @@ constexpr kj::StringPtr operator "" _kj(const char* str, size_t n) {
   return kj::StringPtr(kj::ArrayPtr<const char>(str, n + 1));
 };
 
+constexpr kj::LiteralStringConst operator "" _kjc(const char* str, size_t n) {
+  return kj::LiteralStringConst(kj::ArrayPtr<const char>(str, n + 1));
+};
+
 KJ_END_HEADER
